/* Copyright (c) 2016-2020 MIT 6.005/6.031 course staff, all rights reserved.
 * Redistribution of original or derived work requires permission of course staff.
 */
package edu.mit.eecs.parserlib;

import java.util.List;



/**
 * Represents an immutable parse tree.  Each node represents the application of a production rule
 * from the grammar, "name ::= expr".  The node's name() is the nonterminal on the left side of the production rule,
 * and the node's children() are the production rules that were matched for each nonterminal used in the righthand side (expr).
 * A node also has text() representing the substring matched by this node's subtree, 
 * and start()/end() indicating where that substring is relative to the entire string parsed.
 * <p>
 * If the grammar used {@literal @}skip to skip over some nonterminals automatically, then the skipped subtrees do not appear 
 * among this node's children(), but can be found by using allChildren() or childrenByName().
 * <p>
 * @param <NT> is an Enum type with one value for every nonterminal in the grammar.  See {@link Parser} for more information.
 * 
 * @author 6.005/6.031 course staff
 *
 */
public interface ParseTree<NT extends Enum<NT>> {

    /**
     * Get this node's name.
     * @return the nonterminal corresponding to this node in the parse tree.
     */
    public NT name();

    /**
     * Get this node's children.
     * @return the children of this node, in order, excluding {@literal @}skipped subtrees
     */
    public List<ParseTree<NT>> children();

    /**
     * Get this subtree's text.
     * @return the substring of the entire parsed string that this subtree matched
     */
    public String text();

    /**
     * Get the offset where this subtree starts in the entire parsed string.
     * @return the offset [0,...length-1] in the entire parsed string where this subtree started to match
     */
    public int start();
    
    /**
     * Get the offset where this subtree ends in the entire parsed string.
     * @return the offset [0,...length-1] in the entire parsed string where this subtree ends
     */
    public int end();
    
    /**
     * Get the children that correspond to a particular production rule 
     * @param name Name of the nonterminal corresponding to the desired production rule.
     * @return children that represent matches of name's production rule.
     */
    public List<ParseTree<NT>> childrenByName(NT name);

    /**
     * Get all of this node's children, including {@literal @}skip rules.
     * @return all the children of this node, in order, including {@literal @}skipped subtrees
     */
    public List<ParseTree<NT>> allChildren();

    /**
     * Test if this node is in a subtree that was {@literal @}skipped.
     * @return true iff this node is part of a subtree that was automatically skipped over by
     * a {@literal @}skip rule.
     */
    public boolean isSkipped();
}
