/* Copyright (c) 2016-2020 MIT 6.005/6.031 course staff, all rights reserved.
 * Redistribution of original or derived work requires permission of course staff.
 */
package edu.mit.eecs.parserlib;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.nio.charset.StandardCharsets;

/**
 * A Parser is an immutable object that is able to take a sequence of characters and return a parse tree according
 * to some grammar.
 * <p>
 * Parsers are constructed by calling compile() with a grammar, which might be stored in a 
 * string, in a file, or read from a stream.
 * <p>
 * Once constructed, a Parser object is used by calling parse() on a sequence of characters (represented 
 * as a string or file or stream).  Its result is a ParseTree showing how that string matches the grammar.
 * <p>
 * The type parameter <code>NT</code> should be an Enum type with the same (case-insensitive) names as 
 * the nonterminals in the grammar. This allows nonterminals to be referred to by your Java code with static checking
 * and type safety. For example, if your grammar is:
 * <pre>String sumGrammar = "expression ::= number '+' number ;  number ::= [0-9]+;"</pre>
 * then you should create a nonterminal enum like this:
 * <pre>enum SumGrammar { EXPRESSION, NUMBER };</pre>
 * and then use:
 * <pre>Parser&lt;SumGrammar&gt;.compile(sumGrammar, SumGrammar.EXPRESSION)</pre>
 * to compile it into a parser.
 * <p>
 * The grammar of a grammar is as follows.
 * <pre>
 *   {@literal @}skip whitespaceAndComments {
 *     grammar ::= ( production | skipBlock )+
 *     production ::= nonterminal '::=' union ';'
 *     skipBlock ::= '@skip' nonterminal '{' production* '}'
 *     union :: = concatenation ('|' concatenation)*
 *     concatenation ::= repetition*
 *     repetition ::= unit repeatOperator?
 *     unit ::= nonterminal | terminal | '(' union ')'
 *   }
 *   nonterminal ::= [a-zA-Z_][a-zA-Z_0-9]*
 *   terminal ::= quotedString | characterSet | anyChar | characterClass
 *   quotedString ::= "'" ([^'\r\n\\] | '\\' . )* "'"   // e.g. 'hello', '\'',  '\r\n\t', ''
 *                  | '"' ([^"\r\n\\] | '\\' . )* '"'   // e.g. "world", "\"", "\r\n\t", ""
 *   characterSet ::= '[' ([^\]\r\n\\] | '\\' . )+ ']'   // e.g. [abc], [a-z], [^a-z], [\]], [\r\n\t]
 *   anyChar ::= '.'
 *   repeatOperator ::= [*+?] | '{' ( number | range | upperBound | lowerBound ) '}'
 *   number ::= [0-9]+
 *   range ::= number ',' number
 *   upperBound ::= ',' number
 *   lowerBound ::= number ','
 *   characterClass ::= '\\' [dsw]     // e.g. \d, \s, \w
 *   whitespaceAndComments ::= (whitespace | oneLineComment | blockComment)*
 *   whitespace ::= [ \t\r\n] 
 *   oneLineComment ::= '//' [^\r\n]* [\r\n]+ 
 *   blockComment ::= '/*' [^*]* '*' ([^/]* '*')* '/'
 * </pre>
 * 
 * @param <NT> is an Enum type with one value for every nonterminal in the grammar
 * @author 6.005/6.031 course staff
 *
 */
public interface Parser<NT extends Enum<NT>> {

    public static final String VERSION = "3.2.0";

    /**
     * Compile a Parser from a grammar represented as a string.
     * @param <NT> an Enum type that contains one value for every nonterminal in the grammar.
     * @param grammar the grammar to use
     * @param rootNonterminal the desired root nonterminal in the grammar
     * @return a parser for the given grammar that will start parsing at rootNonterminal.
     * @throws UnableToParseException if the grammar has a syntax error
     */
    public static <NT extends Enum<NT>> Parser<NT> compile(String grammar, NT rootNonterminal) throws UnableToParseException {
        return edu.mit.eecs.parserlib.internal.GrammarCompiler.compile(grammar, rootNonterminal);
    }
    
    /**
     * Compile a Parser from a grammar represented as a Reader stream.
     * @param <NT> an Enum type that contains one value for every nonterminal in the grammar.
     * @param in contains the grammar
     * @param rootNonterminal the desired root nonterminal in the grammar
     * @return a parser for the given grammar that will start parsing at rootNonterminal.
     * @throws UnableToParseException if the grammar has a syntax error
     * @throws IOException if the stream has an I/O error
     */
    public static <NT extends Enum<NT>> Parser<NT> compile(Reader in, NT rootNonterminal) throws UnableToParseException, IOException {
        return edu.mit.eecs.parserlib.internal.GrammarCompiler.compile(in, rootNonterminal);
    }
    
    /**
     * Compile a Parser from a grammar stored in a file.
     * @param <NT> an Enum type that contains one value for every nonterminal in the grammar.
     * @param f file containing the grammar.  Required to have UTF-8 encoding; 
     *           if you need a different encoding, use compile(new FileReader(...),...) 
     *           to choose the encoding yourself instead.
     * @param rootNonterminal the desired root nonterminal in the grammar
     * @return a parser for the given grammar that will start parsing at rootNonterminal.
     * @throws UnableToParseException if the grammar has a syntax error
     * @throws IOException if the file is missing or has an I/O error
     */
    public static <NT extends Enum<NT>> Parser<NT> compile(File f, NT rootNonterminal) throws UnableToParseException, IOException {
        return edu.mit.eecs.parserlib.internal.GrammarCompiler.compile(f, rootNonterminal);
    }

    /**
     * Compile a Parser from a grammar represented as an InputStream.
     * @param <NT> an Enum type that contains one value for every nonterminal in the grammar.
     * @param in stream containing the grammar.  Required to have UTF-8 encoding; if you need a different encoding,
     *           if you need a different encoding, use compile(new InputStreamReader(...),...)
     *           to choose the encoding yourself instead.
     * @param rootNonterminal the desired root nonterminal in the grammar
     * @return a parser for the given grammar that will start parsing at rootNonterminal.
     * @throws UnableToParseException if the grammar has a syntax error
     * @throws IOException if the stream has an I/O error
     */
    public static <NT extends Enum<NT>> Parser<NT> compile(InputStream in, NT rootNonterminal) throws UnableToParseException, IOException {
        return edu.mit.eecs.parserlib.internal.GrammarCompiler.compile(in, rootNonterminal);
    }
    
    /**
     * Parses a string based on the grammar internally represented by the parser.
     * @param string string to parse
     * @return {@link ParseTree} representing a successful parse of the string
     * @throws UnableToParseException if string cannot be parsed, describing approximately where the parsing error occurred
     */
    public ParseTree<NT> parse(String string) throws UnableToParseException;

    /**
     * Parses a stream based on the grammar internally represented by the parser.
     * @param in stream from which to read the text to be parsed.
     * @return {@link ParseTree} representing a successful parse of the content of the stream
     * @throws UnableToParseException if the stream cannot be parsed, describing approximately where the parsing error occurred
     * @throws IOException if the stream has an I/O error.
     */
    public ParseTree<NT> parse(Reader in) throws UnableToParseException, IOException;

    /**
     * Parses a file based on the grammar internally represented by the parser.
     * @param f File containing the text to be parsed. Required to have UTF-8 encoding; 
     *          if you need a different encoding, use parse(new FileReader(...),...) 
     *          to choose the encoding yourself instead.
     * @return {@link ParseTree} representing a successful parse of the content of the file
     * @throws UnableToParseException if the file cannot be parsed, describing approximately where the parsing error occurred
     * @throws IOException if the file has an I/O error.
     */
    public default ParseTree<NT> parse(File f) throws UnableToParseException, IOException {
        return parse(new FileInputStream(f));        
    }

    /**
     * Parses a stream based on the grammar internally represented by the parser.
     * @param stream stream from which to read the text to be parsed.  Required to have UTF-8 encoding; 
     *               if you need a different encoding, use compile(new InputStreamReader(...),...) 
     *               to choose the encoding yourself instead.
     * @return {@link ParseTree} representing a successful parse of the content of the stream
     * @throws UnableToParseException if the stream cannot be parsed, describing approximately where the parsing error occurred
     * @throws IOException if the stream has an I/O error.
     */
    public default ParseTree<NT> parse(InputStream stream) throws UnableToParseException, IOException {
        return parse(new InputStreamReader(stream, StandardCharsets.UTF_8));
    }
    

}
