/* Copyright (c) 2016-2020 MIT 6.005/6.031 course staff, all rights reserved.
 * Redistribution of original or derived work requires permission of course staff.
 */
package edu.mit.eecs.parserlib;

import java.awt.Desktop;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URLEncoder;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

/**
 * Support for visualizing a parse tree in a web browser, for debugging purposes.
 * 
 * @author 6.005/6.031 course staff
 */
public class Visualizer {
    private Visualizer() {} // prevents Java's default constructor from appearing in the Javadoc
    
    private static final int MAX_URL_LENGTH_FOR_DESKTOP_BROWSE = 2020;
    
    /**
     * Tries to show a visualization of a parse tree in your web browser. 
     * If you are not connected to the internet, or if the method cannot launch the browser, 
     * it will print a very long URL to your console which you can then copy to your browser to 
     * see the visualization.
     * @param parseTree tree to visualize
     * @param <NT> the enumeration of symbols in the parse tree's grammar
     */
    public static <NT extends Enum<NT>> void showInBrowser(ParseTree<NT> parseTree) {
        String base = "http://6031.mit.edu/parserlib/" + Parser.VERSION + "/visualizer.html";
        List<ParseTree<NT>> empty = Collections.emptyList();
        String code = forDisplay(parseTree, empty.iterator(), parseTree);
        try {
            String url = base + "?code=" + URLEncoder.encode(code, "UTF-8");
            if(Desktop.isDesktopSupported() && url.length() < MAX_URL_LENGTH_FOR_DESKTOP_BROWSE)
            {
                Desktop desktop = Desktop.getDesktop();
                if(desktop.isSupported(Desktop.Action.BROWSE)){
                    try{
                        URI uri = new URI(url);
                        desktop.browse(uri);
                        return;
                    } catch (URISyntaxException e) {
                        throw new RuntimeException(e);
                    } catch(IOException e){
                        // fall through to fallback below
                    }
                }
            }
            
            // desktop.browse() didn't work, so provide manual instructions
            System.out.println("Go to " + base + "\nand copy and paste this code into the textbox:\n" + code);
            
        } catch (UnsupportedEncodingException e) {
            throw new RuntimeException("Tried using UTF-8 encoding but it is not supported.");
        }                
    }

    /**
     * Generate an HTML visualization of the parse tree and write it to a file.
     * @param parseTree tree to visualize
     * @param filename output file. Advisable to give it ".html" extension, though not required.
     * @throws IOException if unable to write the file.
     * @param <NT> the enumeration of symbols in the parse tree's grammar
     */
    public static <NT extends Enum<NT>> void writeToFile(ParseTree<NT> parseTree, String filename) throws IOException {
        try {
            PrintWriter out = new PrintWriter(filename, "UTF-8");
            InputStream in = Visualizer.class.getResourceAsStream("/visualizer.html"); 
            BufferedReader reader = new BufferedReader(new InputStreamReader(in));
            String line = reader.readLine();
            while(line != null){
                if(!line.equals("//CODEHERE")){
                    out.println(line);
                }else{
                    List<ParseTree<NT>> empty = Collections.emptyList();
                    String code = forDisplay(parseTree, empty.iterator(), parseTree);
                    out.println("return '" + URLEncoder.encode(code, "UTF-8") + "';");
                }
                line = reader.readLine();
            }
            out.flush();
            out.close();
        } catch (UnsupportedEncodingException e) {
            throw new RuntimeException("Tried using UTF-8 encoding but it is not supported.");
        }        
    }
    
    private static <NT extends Enum<NT>> String forDisplay(final ParseTree<NT> node, final Iterator<ParseTree<NT>> sibilings, final ParseTree<NT> parent) {
        final String name =  node.name().toString().toLowerCase();
        String s = "nd(";
        if (node.children().isEmpty()){
            s += "\"" + name + "\",nd(\"'"+ cleanString(node.text()) + "'\"),";
        } else {
            s += "\"" + name + "\",";
            Iterator<ParseTree<NT>> it = node.allChildren().iterator();
            final ParseTree<NT> firstChild = it.next();
            String childrenExpression = forDisplay(firstChild, it, node);
            if (node.start() < firstChild.start()) {
                // node and its first child have a gap between them that must have been matched as a terminal
                // in the rule for node.  Insert it as a quoted string.
                childrenExpression = precedeByTerminal(node.text().substring(0, firstChild.start() - node.start()), childrenExpression);                
            }
            s += childrenExpression + ",";
        }
        if(sibilings.hasNext()){
            final ParseTree<NT> sibling = sibilings.next();
            String siblingExpression = forDisplay(sibling, sibilings, parent);
            if (node.end() < sibling.start()) {
                // node and its sibling have a gap between them that must have been matched as a terminal
                // in the rule for parent.  Insert it as a quoted string.
                siblingExpression = precedeByTerminal(parent.text().substring(node.end() - parent.start(), sibling.start() - parent.start()), siblingExpression);
            }
            s += siblingExpression;
        }else{
            String siblingExpression = "uu";
            if (node.end() < parent.end()) {
                // There's a gap between the end of node and the end of its parent, which must be a terminal matched by parent.
                // Insert it as a quoted string.
                siblingExpression = precedeByTerminal(parent.text().substring(node.end() - parent.start()), siblingExpression);
            }
            s += siblingExpression;
        }
        if (node.isSkipped()) {
            s += ",true";
        }
        s += ")";
        return s;
    }
    
    private static String precedeByTerminal(String terminal, String expression) {
        return "nd(\"'"+ cleanString(terminal) + "'\", uu, " + expression + ")";     
    }

    private static String cleanString(String s){
        String rvalue = s.replace("\\", "\\\\");
        rvalue = rvalue.replace("\"", "\\\"");
        rvalue = rvalue.replace("\n", "\\n");
        rvalue = rvalue.replace("\r", "\\r");
        return rvalue;
    }
    
}
