/* Copyright (c) 2016-2020 MIT 6.005/6.031 course staff, all rights reserved.
 * Redistribution of original or derived work requires permission of course staff.
 */
package edu.mit.eecs.parserlib.internal;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import edu.mit.eecs.parserlib.UnableToParseException;


/**
 * 
 * This class represents a choice between multiple terms in the grammar, corresponding to the
 * alternation operator <code>(a | b)</code> that matches either <code>a</code> or <code>b</code>.
 * The type parameter <code>NT</code> is an Enum type that represents the set of nonterminals available in the grammar.
 * @author asolar
 *
 * @param <NT>
 */
public class Choice<NT extends Enum<NT>> extends GrammarTerm<NT> {
    private final List<GrammarTerm<NT>> choices;
    
    /**
     * Choice between two or more terms in the grammar. 
     * Class keeps an unmodifiable reference to the list, but other
     * classes should not modify the choices list.
     * @param choices list of choices.
     */
    public Choice(List<GrammarTerm<NT>> choices) {
        super();        
        this.choices = Collections.unmodifiableList(choices);
    }

    @Override
    public ParseResult<NT> parse(String s, int pos, Map<NT, GrammarTerm<NT>> definitions, ParserState<NT> state) throws UnableToParseException {
        List<ParseResult<NT>> results = new ArrayList<>();
        int deepestPosReachedByAFailure = 0;
        String expectedTextAtDeepestPosReached = "";
        for(GrammarTerm<NT> choice : choices){
            ParseResult<NT> pr = choice.parse(s, pos, definitions, state);
            if(!pr.failed()){
                results.add(pr);
            } else {
                if (pr.pos > deepestPosReachedByAFailure) {
                    deepestPosReachedByAFailure = pr.pos;
                    expectedTextAtDeepestPosReached = pr.expectedText;
                } else if (pr.pos == deepestPosReachedByAFailure) {
                    expectedTextAtDeepestPosReached += (expectedTextAtDeepestPosReached.isEmpty() ? "" : "|") + pr.expectedText;
                }
            }
        }
        if(results.isEmpty()){
            return new ParseResult<NT>(state.current(), expectedTextAtDeepestPosReached, deepestPosReachedByAFailure);
        }
        Optional<ParseResult<NT>> result = Optional.empty();        
        for(ParseResult<NT> current : results){
            if(result.isPresent()){
                if(result.get().pos < current.pos){
                    result = Optional.of(current);
                }
            }else{
                result = Optional.of(current);
            }
        }
        
        ParseResult<NT> success = result.get();
        if (deepestPosReachedByAFailure > success.pos) {
            // we could have matched more -- remember where we failed
            success = success.addLastFailure(new ParseResult<NT>(state.current(), expectedTextAtDeepestPosReached, deepestPosReachedByAFailure));
        }
        return success;
    }
    
    @Override
    public String toString(){
        String s = "";
        boolean first = true;
        for(GrammarTerm<NT> choice: choices){
            if(!first){
                s += "|";
            }
            first = false;
            s += choice;
        }
        return "(" + s + ")";
    }
    
}
