/* Copyright (c) 2016-2020 MIT 6.005/6.031 course staff, all rights reserved.
 * Redistribution of original or derived work requires permission of course staff.
 */
package edu.mit.eecs.parserlib.internal;

import java.util.Collections;
import java.util.List;
import java.util.Map;

import edu.mit.eecs.parserlib.UnableToParseException;


/**
 * This class represents a concatenation of multiple terms in the grammar.
 * The type parameter <code>NT</code> is an Enum type that represents the set of nonterminals available in the grammar.
 * @author asolar
 *
 * @param <NT>
 */
public class Concatenation<NT extends Enum<NT>> extends GrammarTerm<NT> {
    private final List<GrammarTerm<NT>> parts;
    
    
    /**
     * Concatenation of two or more terms in the grammar. 
     * Class keeps an unmodifiable reference to the list, but other
     * classes should not modify the parts list.
     * @param parts list of terms to concatenate.
     */
    public Concatenation(List<GrammarTerm<NT>> parts) {
        super();
        this.parts = Collections.unmodifiableList(parts);
    }
    
    @Override
    public ParseResult<NT> parse(String s, int pos, Map<NT, GrammarTerm<NT>> definitions, ParserState<NT> state) throws UnableToParseException {
        ParseResult<NT> prout=null;
        for(GrammarTerm<NT> gt: parts){
            ParseResult<NT> pr = gt.parse(s, pos, definitions, state);
            if(pr.failed())
                return pr;
            pos = pr.pos;
            if(prout != null){
                prout = prout.mergeResult(pr);
            }else{
                prout = pr;
            }
        }
        return prout;
    }
    
    @Override
    public String toString(){
        String s = "";        
        for(GrammarTerm<NT> choice: parts){            
            s += choice;
        }
        return "(" + s + ")";
    }

}
