/* Copyright (c) 2016-2020 MIT 6.005/6.031 course staff, all rights reserved.
 * Redistribution of original or derived work requires permission of course staff.
 */
package edu.mit.eecs.parserlib.internal;

import java.util.ArrayList;
import java.util.List;

/**
 * Utility functions for making human-readable strings.
 */
public class Display {

    private static final String[][] ESCAPES = {
            { "\n", "\u2424", "newline" },
            { "\r", "\u240D", "carriage return" },
            { "\t", "\u21E5", "tab" },
    };
    
    /**
     * Replace common unprintable characters by their escape codes, for human reading.
     * Should be idempotent, i.e. if x = escapeForReading(y), then x.equals(escapeForReading(x)).
     * @param string to escape
     * @param quote quotes to put around string, or "" if no quotes required
     * @return string with escape codes replaced, preceded and followed by quote, with a human-readable legend appended to the end
     *         explaining what the replacement characters mean.
     */
    public static String escapeForReading(String string, String quote) {
        String result = string;
        final List<String> legend = new ArrayList<>();

        for (final String[] escape : ESCAPES) {
            final String unprintableChar = escape[0];
            final String humanReadableVersion = escape[1];
            final String description = escape[2];
            if (result.contains(unprintableChar)) {
                result = result.replace(unprintableChar, humanReadableVersion);
                legend.add(humanReadableVersion + " means " + description);
            }
        }
        
        result = quote + result + quote;
        if (!legend.isEmpty()) {
            result += " (where " + String.join(", ", legend) + ")";
        }
        return result;
    }
    
    /**
     * @param string to shorten
     * @param pos offset in string, 0<=pos<string.length()
     * @return a short snippet of the part of string starting at offset pos,
     * in human-readable form
     */
    public static String snippet(String string, int pos) {
        final int maxCharsToShow = 10;
        int end = Math.min(pos + maxCharsToShow, string.length());
        String result = string.substring(pos, end) + (end < string.length() ? "..." : "");
        if (result.isEmpty()) result = "end of string";
        return escapeForReading(result, "");
    }

    /**
     * Translates a string offset into a line number.
     * @param string in which offset occurs
     * @param pos offset in string, 0<=pos<string.length()
     * @return the 1-based line number of the character at offset pos in string,
     * as if string were being viewed in a text editor
     */
    public static int toLine(String string, int pos) {
        int lineCount = 1;
        for (int newline = string.indexOf('\n'); newline != -1 && newline < pos; newline = string.indexOf('\n', newline+1)) {
            ++lineCount;
        }
        return lineCount;
    }

    /**
     * Translates a string offset into a column number.
     * @param string in which offset occurs
     * @param pos offset in string, 0<=pos<string.length()
     * @return the 1-based column number of the character at offset pos in string,
     * as if string were being viewed in a text editor with tab size 1 (i.e. a tab is treated like a space)
     */
    public static int toColumn(String string, int pos) {
        int lastNewlineBeforePos = string.lastIndexOf('\n', pos-1);
        int totalSizeOfPrecedingLines = (lastNewlineBeforePos != -1) ? lastNewlineBeforePos+1 : 0;
        return pos - totalSizeOfPrecedingLines + 1;
    }

    /**
     * @param string to describe
     * @param pos offset in string, 0<=pos<string.length()
     * @return a human-readable description of the location of the character at offset pos in string
     * (using offset and/or line/column if appropriate)
     */
    public static String describeLocation(String string, int pos) {
        String result = "offset " + pos;
        if (string.indexOf('\n') != -1) {
            result += " (line " + toLine(string, pos) + " column " + toColumn(string, pos) +")";
        }
        return result;
    }

    /**
     * Make a human-readable error message explaining a parse error and where it was found in the input.
     * @param message brief message stating what error occurred
     * @param nonterminalName name of deepest nonterminal that parser was trying to match when parse failed
     * @param expectedText human-readable description of what string literals the parser was expecting there;
     *            e.g. ";", "[ \r\n\t]", "1|2|3"
     * @param stringBeingParsed original input to parse()
     * @param pos offset in stringBeingParsed where error occurred
     * @param nameOfStringBeingParsed human-readable description of where stringBeingParsed came from;
     *             e.g. "grammar" if stringBeingParsed was the input to Parser.compile(),
     *             or "string being parsed" if stringBeingParsed was the input to Parser.parse()
     * @return a multiline human-readable message that states the error, its location in the input,
     *         what text was expected and what text was actually found.
     */
    public static String makeErrorMessage(String message, String nonterminalName, String expectedText, String stringBeingParsed, int pos, String nameOfStringBeingParsed) {
        String result = message;
        if (!result.isEmpty()) result += "\n";
        result += "Error at " + Display.describeLocation(stringBeingParsed, pos) + " of "  + nameOfStringBeingParsed + "\n" 
                + "  trying to match " + nonterminalName.toUpperCase() + "\n"
                + "  expected " + Display.escapeForReading(expectedText, "")
                + (stringBeingParsed != null && !stringBeingParsed.isEmpty()
                    ? "\n   but saw " + Display.snippet(stringBeingParsed, pos)
                    : "");
        return result;
    }

    /**
     * Indent a multi-line string by preceding each line with prefix.
     * @param string string to indent
     * @param prefix prefix to use for indenting
     * @return string with prefix inserted at the start of each line
     */
    public static String indent(String string, String prefix) {
        String result = "";
        int charsCopied = 0;
        do {
            int newline = string.indexOf('\n', charsCopied);
            int endOfLine = newline != -1 ? newline+1 : string.length();
            result += prefix + string.substring(charsCopied, endOfLine);
            charsCopied = endOfLine;
        } while (charsCopied < string.length());
        return result;
    }
}
