/* Copyright (c) 2016-2020 MIT 6.005/6.031 course staff, all rights reserved.
 * Redistribution of original or derived work requires permission of course staff.
 */
package edu.mit.eecs.parserlib.internal;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import edu.mit.eecs.parserlib.ParseTree;
import edu.mit.eecs.parserlib.Parser;
import edu.mit.eecs.parserlib.UnableToParseException;

/**
 * Convert a grammar into a Parser.
 */
public class GrammarCompiler {

    enum Terms {GRAMMAR, PRODUCTION, SKIPBLOCK, UNION, CONCATENATION, 
        REPETITION, REPEATOPERATOR, UNIT, NONTERMINAL, TERMINAL, QUOTEDSTRING, 
        NUMBER, RANGE, UPPERBOUND, LOWERBOUND,
        CHARACTERSET, ANYCHAR, CHARACTERCLASS, WHITESPACE, ONELINECOMMENT, BLOCKCOMMENT, SKIP}
    
    private static GrammarConstruct<Terms> grammarGrammar = new GrammarConstruct<Terms>() {
        @Override
        public Map<Terms, GrammarTerm<Terms>> get() {

            // grammar ::= ( production | skipBlock )+
            grammar.put(Terms.GRAMMAR, cat(nt(Terms.SKIP), star(cat(or (nt(Terms.PRODUCTION), nt(Terms.SKIPBLOCK)), nt(Terms.SKIP)))));

            // skipBlock ::= '@skip' nonterminal '{' production* '}'
            grammar.put(Terms.SKIPBLOCK, cat(str("@skip"), nt(Terms.SKIP), failfast(nt(Terms.NONTERMINAL)), nt(Terms.SKIP), 
                                         str('{'), 
                                         failfast(cat(nt(Terms.SKIP), 
                                                      star(cat(nt(Terms.PRODUCTION), nt(Terms.SKIP))),
                                                      str('}')))));

            // production ::= nonterminal '::=' union ';'
            grammar.put(Terms.PRODUCTION, cat(nt(Terms.NONTERMINAL), nt(Terms.SKIP), 
                                          str("::="), 
                                          failfast(cat(nt(Terms.SKIP), 
                                                   nt(Terms.UNION), nt(Terms.SKIP), 
                                                   str(';')))));

            // union :: = concatenation ('|' concatenation)*
            grammar.put(Terms.UNION, cat(nt(Terms.CONCATENATION), star(cat(nt(Terms.SKIP), str('|'), nt(Terms.SKIP), nt(Terms.CONCATENATION)))));
            
            // concatenation :: = repetition* 
            grammar.put(Terms.CONCATENATION, star(cat(nt(Terms.REPETITION), nt(Terms.SKIP))));
            
            // repetition ::= unit repeatOperator?
            grammar.put(Terms.REPETITION, cat(nt(Terms.UNIT), nt(Terms.SKIP), option(nt(Terms.REPEATOPERATOR))));

            // repeatOperator ::= [*+?] | '{' ( number | range | upperBound | lowerBound ) '}'
            grammar.put(Terms.REPEATOPERATOR, or(regex("[*+?]"), 
                                                 cat(str("{"),
                                                     or(nt(Terms.NUMBER),
                                                        nt(Terms.RANGE),
                                                        nt(Terms.UPPERBOUND),
                                                        nt(Terms.LOWERBOUND)),
                                                     str("}"))));
            // number ::= [0-9]+
            grammar.put(Terms.NUMBER, plus(regex("[0-9]")));
            
            // range ::= number ',' number
            grammar.put(Terms.RANGE, cat(nt(Terms.NUMBER), str(","), nt(Terms.NUMBER)));
            
            // upperBound ::= ',' number
            grammar.put(Terms.UPPERBOUND, cat(str(","), nt(Terms.NUMBER)));

            // lowerBound ::= number ','
            grammar.put(Terms.LOWERBOUND, cat(nt(Terms.NUMBER), str(",")));

            // unit ::= nonterminal | terminal | '(' union ')'
            grammar.put(Terms.UNIT, or(nt(Terms.NONTERMINAL),
                                   nt(Terms.TERMINAL),
                                   cat(str('('), nt(Terms.SKIP), nt(Terms.UNION), nt(Terms.SKIP), str(')'))));

            // nonterminal ::= [a-zA-Z_][a-zA-Z_0-9]*
            grammar.put(Terms.NONTERMINAL, cat(regex("[a-zA-Z_]"), star(regex("[a-zA-Z_0-9]*"))));

            // terminal ::= quotedString | characterSet | anyChar | characterClass
            grammar.put(Terms.TERMINAL, or(nt(Terms.QUOTEDSTRING), nt(Terms.CHARACTERSET), nt(Terms.ANYCHAR), nt(Terms.CHARACTERCLASS)));
            
            // quotedString ::= "'" ([^'\r\n\\] | '\\' . )* "'" | '"' ([^"\r\n\\] | '\\' . )* '"'
            grammar.put(Terms.QUOTEDSTRING, or(cat(str("'"), failfast(star(or(regex("[^'\r\n\\\\]"),
                                                                           cat(str('\\'), regex("."))))),
                                                   str("'")),
                                               cat(str('"'), failfast(star(or(regex("[^\"\r\n\\\\]"),
                                                                              cat(str('\\'), regex("."))))), 
                                                   str('"'))));

            // characterSet ::= '[' ([^\]\r\n\\] | '\\' . )+ ']'
            grammar.put(Terms.CHARACTERSET, cat(str('['), 
                                            failfast(cat(plus(or(regex("[^\\]\r\n\\\\]"),
                                                                 cat(str('\\'), regex(".")))))), 
                                            str(']')));
            
            // anyChar ::= '.'
            grammar.put(Terms.ANYCHAR, str('.'));

            // characterClass ::= '\\' [dsw]
            grammar.put(Terms.CHARACTERCLASS, cat(str('\\'), failfast(regex("[dsw]"))));
            
            // whitespace ::= [ \t\r\n]
            grammar.put(Terms.WHITESPACE, regex("[ \t\r\n]"));
            grammar.put(Terms.ONELINECOMMENT, cat(str("//"), star(regex("[^\r\n]")), or(str("\r\n"), str('\n'), str('\r'))));
            grammar.put(Terms.BLOCKCOMMENT, cat(str("/*"),
                                            cat(star(regex("[^*]")), str('*')),
                                            star(cat(regex("[^/]"), star(regex("[^*]")), str('*'))), 
                                            str('/')));
            grammar.put(Terms.SKIP, star(or(nt(Terms.WHITESPACE), nt(Terms.ONELINECOMMENT), nt(Terms.BLOCKCOMMENT))));
            
            return grammar;
        }            
    };
    
    
    private static <NT extends Enum<NT>> Parser<NT> compile(ParseTree<Terms> grammarTree, NT rootNonterminal) throws UnableToParseException{
        return new edu.mit.eecs.parserlib.internal.InternalParser<NT>(new MyGrammarConstruct<NT>(grammarTree, rootNonterminal), rootNonterminal);
    }
    
    private static class MyGrammarConstruct<NT extends Enum<NT>> extends GrammarConstruct<NT> {
        private final ParseTree<Terms> grammarTree;
        private final NT rootNonterminal;
        
        public MyGrammarConstruct(ParseTree<Terms> grammarTree, NT rootNonterminal) {
            this.grammarTree = grammarTree;
            this.rootNonterminal = rootNonterminal;
        }
        
        /**
         * Returns an Enum value whose name is equal to name.
         * Method is case insensitive.
         * @param name name of the Enum element.
         * @return a value of the NT Enum type.
         */
        @SuppressWarnings("unchecked")
        NT valueOf(String name) throws UnableToParseException {
            // This could have been a call to Enum.valueOf(rootNonterminal.getClass(), name);
            // but we want it to be case insensitive.                
            for(Object e : java.util.EnumSet.allOf(rootNonterminal.getClass())){            
                if(e.toString().toLowerCase().equals(name.toLowerCase())){
                    return (NT)e;
                }
            }
            throw new GrammarException("nonterminal " + name.toUpperCase() + " is used in the grammar but not found in the enum type " 
                                       + rootNonterminal.getClass().getName());
        }
        
        @Override
        public Map<NT, GrammarTerm<NT>> get() throws UnableToParseException {
            // productions outside @skip blocks
            makeProductions(grammarTree.childrenByName(Terms.PRODUCTION), Optional.empty());
            
            // production inside @skip blocks
            for (ParseTree<Terms> skipBlock : grammarTree.childrenByName(Terms.SKIPBLOCK)) {
                String nonterminal = skipBlock.childrenByName(Terms.NONTERMINAL).get(0).text();
                makeProductions(skipBlock.childrenByName(Terms.PRODUCTION), Optional.of(valueOf(nonterminal)));
            }
            
            if(!grammar.containsKey(rootNonterminal)){
                throw new GrammarException("grammar is missing a definition for the root nonterminal " + rootNonterminal);
            }
            return grammar;
        }

        private void makeProductions(List<ParseTree<Terms>> productions, Optional<NT> skip) throws UnableToParseException {
            for (ParseTree<Terms> production : productions) {
                String nonterminal = production.childrenByName(Terms.NONTERMINAL).get(0).text();
                GrammarTerm<NT> expression = makeGrammarTerm(production.childrenByName(Terms.UNION).get(0), skip);
                if (skip.isPresent()) expression = cat(makeSkip(skip.get()), expression, makeSkip(skip.get()));
                
                NT s = valueOf(nonterminal);
                if (grammar.containsKey(s)) {
                    // grammar already has a production for s; or expression onto it
                    grammar.put(s, or(grammar.get(s), expression));
                } else {
                    grammar.put(s, expression);
                }
            }                
        }
        
        private GrammarTerm<NT> makeGrammarTerm(ParseTree<Terms> tree, Optional<NT> skip) throws UnableToParseException {
            Terms rule = tree.name();
            if (rule  == Terms.UNION) {
                Iterator<ParseTree<Terms>> children = tree.childrenByName(Terms.CONCATENATION).iterator();
                // union's grammar rule guarantees at least one child
                assert children.hasNext(); 
                // start with first child
                GrammarTerm<NT> result = makeGrammarTerm(children.next(), skip); 
                while (children.hasNext()) {
                    // or on remaining children
                    result = or(result, makeGrammarTerm(children.next(), skip));
                }
                return result;
            } else if (rule==Terms.CONCATENATION) {
                Iterator<ParseTree<Terms>> children = tree.childrenByName(Terms.REPETITION).iterator();
                if (!children.hasNext()) {
                    // use empty string to represent an empty concatenation
                    return str("");
                } else {
                    // start with first child
                    GrammarTerm<NT> result = makeGrammarTerm(children.next(), skip);
                    while (children.hasNext()) {
                        // concatenate remaining children
                        if (skip.isPresent()) result = cat(result, makeSkip(skip.get()));
                        result = cat(result, makeGrammarTerm(children.next(), skip));
                    }
                    return result;
                }
            } else if (rule==Terms.REPETITION) {
                GrammarTerm<NT> result = makeGrammarTerm(tree.childrenByName(Terms.UNIT).get(0), skip);
                if (!tree.childrenByName(Terms.REPEATOPERATOR).isEmpty()) {
                    if (skip.isPresent()) result = cat(result, makeSkip(skip.get()));
                    ParseTree<Terms> op = tree.childrenByName(Terms.REPEATOPERATOR).get(0);
                    if(op.text().equals("*")){
                        result = star(result);
                    }
                    if(op.text().equals("+")){
                        result = plus(result);
                    }
                    if(op.text().equals("?")){
                        result = option(result);
                    }
                    if (op.children().size() > 0) {
                        // {n,m} repetition
                        ParseTree<Terms> opChild = op.children().get(0);
                        Repetition.HowMany howmany;
                        switch (opChild.name()) {
                        case NUMBER:
                        {
                            int n = Integer.valueOf(opChild.text());
                            howmany = new Repetition.Between(n, n);
                            break;
                        }
                        case RANGE:
                        {
                            int n = Integer.valueOf(opChild.children().get(0).text());
                            int m = Integer.valueOf(opChild.children().get(1).text());
                            howmany = new Repetition.Between(n, m);
                            break;
                        }
                        case UPPERBOUND:
                        {
                            int m = Integer.valueOf(opChild.children().get(0).text());
                            howmany = new Repetition.Between(0, m);
                            break;
                        }
                        case LOWERBOUND:
                        {
                            int n = Integer.valueOf(opChild.children().get(0).text());
                            howmany = new Repetition.AtLeast(n);
                            break;
                        }
                        default:
                            throw new AssertionError("internal error: unknown range: " + opChild.name());
                        }
                        result = repeat(result, howmany);
                    }
                }
                return result;
            } else if (rule==Terms.UNIT) {
                if (!tree.childrenByName(Terms.NONTERMINAL).isEmpty()) {
                    return makeGrammarTerm(tree.childrenByName(Terms.NONTERMINAL).get(0), skip);
                } else if (!tree.childrenByName(Terms.TERMINAL).isEmpty()) {
                    return makeGrammarTerm(tree.childrenByName(Terms.TERMINAL).get(0), skip);
                } else if (!tree.childrenByName(Terms.UNION).isEmpty()) {
                    return makeGrammarTerm(tree.childrenByName(Terms.UNION).get(0), skip);
                } else {
                    throw new AssertionError("internal error: no children found");
                }
            } else if (rule==Terms.NONTERMINAL) {
                return nt(valueOf(tree.text()));
            } else if (rule==Terms.TERMINAL) {
                String content = tree.text();
                if (!tree.childrenByName(Terms.QUOTEDSTRING).isEmpty()) {
                    String literal = replaceEscapeSequences(content.substring(1, content.length()-1));
                    return str(literal);
                } else if (!tree.childrenByName(Terms.CHARACTERSET).isEmpty()) {
                    return regex(content);                        
                } else if (!tree.childrenByName(Terms.ANYCHAR).isEmpty()
                           || !tree.childrenByName(Terms.CHARACTERCLASS).isEmpty()) {
                    // these expression correspond directly to one-character regexes:
                    // .   \d  \s  \w
                    return regex(content);
                } else {
                    throw new AssertionError("internal error: no children found");
                }
            } else {
                throw new AssertionError("internal error: unknown grammar rule: " + rule);
            }
        }

        /**
         * @param nonterminal nonterminal in @skip clause
         * @return a grammar term for skipping over occurrences of it
         */
        private GrammarTerm<NT> makeSkip(NT nonterminal) {
            return skip(nonterminal);
        }
        
        /**
         * Replace \t, \r, \n with their character codes.
         * Replaces all other \x with literal x.
         */
        private static String replaceEscapeSequences(String s) {
            Matcher m = Pattern.compile("\\\\(.)").matcher(s);
            StringBuffer sb = new StringBuffer();
            while (m.find()) {
                char oldChar = m.group(1).charAt(0);
                char newChar;
                switch (oldChar) {
                case 't': newChar = '\t'; break;
                case 'r': newChar = '\r'; break;
                case 'n': newChar = '\n'; break;
                default:  newChar = oldChar; break;
                }
                // The normal .find()/.appendReplacement()/.appendTail() idiom here would 
                // be m.appendReplacement(sb, newChar), but we can't do that because 
                // appendReplacement() treats backslashes and dollar signs specially.
                // We just want the literal newChar, so give appendReplacement nothing,
                // and append it ourselves.
                m.appendReplacement(sb, "");
                sb.append(newChar);
            }
            m.appendTail(sb);
            return sb.toString();
        }

    }
    
    
    /**
     * @see Parser#compile(String,Enum)
     */
    public static <NT extends Enum<NT>> Parser<NT> compile(String grammar, NT rootNonterminal) throws UnableToParseException {
        try {
            Parser<Terms> grammarParser = new edu.mit.eecs.parserlib.internal.InternalParser<Terms>(grammarGrammar, Terms.GRAMMAR);
            ParseTree<Terms> grammarTree = grammarParser.parse(grammar);
            return compile(grammarTree, rootNonterminal);
        } catch (ParseException e) {
            throw translateToGrammarException(e);
        }
    }
    
    /**
     * @see Parser#compile(Reader,Enum)
     */
    public static <NT extends Enum<NT>> Parser<NT> compile(Reader in, NT rootNonterminal) throws UnableToParseException, IOException {
        try {
            Parser<Terms> grammarParser = new edu.mit.eecs.parserlib.internal.InternalParser<Terms>(grammarGrammar, Terms.GRAMMAR);
            ParseTree<Terms> grammarTree = grammarParser.parse(in);
            return compile(grammarTree, rootNonterminal);
        } catch (ParseException e) {
            throw translateToGrammarException(e);
        }
    }
    
    /**
     * @see Parser#compile(File,Enum)
     */
    public static <NT extends Enum<NT>> Parser<NT> compile(File in, NT rootNonterminal) throws UnableToParseException, IOException {
        try {
            if(!in.canRead()){
                throw new IOException("Cannot read from file " + in.getAbsolutePath());
            }
            Parser<Terms> grammarParser = new edu.mit.eecs.parserlib.internal.InternalParser<Terms>(grammarGrammar, Terms.GRAMMAR);
            ParseTree<Terms> grammarTree = grammarParser.parse(in);
            return compile(grammarTree, rootNonterminal);
        } catch (ParseException e) {
            throw translateToGrammarException(e);
        }
    }

    /**
     * @see Parser#compile(InputStream,Enum)
     */
    public static <NT extends Enum<NT>> Parser<NT> compile(InputStream in, NT rootNonterminal) throws UnableToParseException, IOException {
        try {
            Parser<Terms> grammarParser = new edu.mit.eecs.parserlib.internal.InternalParser<Terms>(grammarGrammar, Terms.GRAMMAR);
            ParseTree<Terms> grammarTree = grammarParser.parse(in);            
            return compile(grammarTree, rootNonterminal);
        } catch (ParseException e) {
            throw translateToGrammarException(e);
        }
    }
    
    private static GrammarException translateToGrammarException(ParseException e) {
        return new GrammarException(Display.makeErrorMessage("grammar doesn't compile",
                                                              e.nonterminalName,
                                                              e.expectedText,
                                                              e.textBeingParsed,
                                                              e.pos,
                                                              "grammar"));
    }
}
