/* Copyright (c) 2016-2020 MIT 6.005/6.031 course staff, all rights reserved.
 * Redistribution of original or derived work requires permission of course staff.
 */
package edu.mit.eecs.parserlib.internal;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import edu.mit.eecs.parserlib.UnableToParseException;

/**
 * Helper class to facilitate instantiating grammars through code.
 * Helper methods nt, option, or, plus, regex, star, and str create GrammarTerms more concisely than calling constructors directly. 
 * 
 * You need to override the <code>get</code> method to make it produce your own grammar.
 * @author asolar
 *
 */
public abstract class GrammarConstruct<NT extends Enum<NT>> {
    
    protected Map<NT, GrammarTerm<NT>> grammar = new HashMap<NT, GrammarTerm<NT>>();
    
    @SafeVarargs
    protected final GrammarTerm<NT> or(GrammarTerm<NT> ... choices){
        List<GrammarTerm<NT>> p = new ArrayList<GrammarTerm<NT>>();
        for(GrammarTerm<NT> gt : choices){
            p.add(gt);
        }
        return new Choice<NT>(p);
    }
    
    @SafeVarargs    
    protected final GrammarTerm<NT> cat(GrammarTerm<NT> ... terms){
        List<GrammarTerm<NT>> p = new ArrayList<GrammarTerm<NT>>();
        for(GrammarTerm<NT> gt : terms){
            p.add(gt);
        }
        return new Concatenation<NT>(p);
    }
    
    protected GrammarTerm<NT> str(String s){
        return new Constant<NT>(s);
    }

    protected GrammarTerm<NT> str(char c){
        return str(String.valueOf(c));
    }
    
    protected GrammarTerm<NT> regex(String regex) {
        return new OneCharacterRegex<NT>(regex);
    }
    
    protected GrammarTerm<NT> nt(NT s){
        return new NonTerminal<NT>(s);
    }
    
    protected GrammarTerm<NT> star(GrammarTerm<NT> gt){
        return new Repetition<NT>(gt, Repetition.HowMany.ZERO_OR_MORE);
    }
    
    protected GrammarTerm<NT> plus(GrammarTerm<NT> gt){
        return new Repetition<NT>(gt, Repetition.HowMany.ONE_OR_MORE);
    }
    
    protected GrammarTerm<NT> option(GrammarTerm<NT> gt){
        return new Repetition<NT>(gt, Repetition.HowMany.ZERO_OR_ONE);
    }

    protected GrammarTerm<NT> repeat(GrammarTerm<NT> gt, Repetition.HowMany howmany){
        return new Repetition<NT>(gt, howmany);
    }
    
    protected GrammarTerm<NT> skip(NT s) {
        return new Skip<NT>(nt(s));
    }
    
    protected GrammarTerm<NT> failfast(GrammarTerm<NT> child) {
        return new FailFast<NT>(child);
    }
    
    /**
     * A particular use of the GrammarConstruct class overrides this method to actually return 
     * a map of GrammarTerms.
     * @return Map of nonterminals and their corresponding grammar terms. All elements in the NT Enum must
     * be represented in this map.
     * @throws UnableToParseException if the nonterminals in the grammar are not in 1-1 correspondence with the symbols in NT
     */
    public abstract Map<NT, GrammarTerm<NT>> get() throws UnableToParseException;

}
