/* Copyright (c) 2016-2020 MIT 6.005/6.031 course staff, all rights reserved.
 * Redistribution of original or derived work requires permission of course staff.
 */
package edu.mit.eecs.parserlib.internal;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import edu.mit.eecs.parserlib.ParseTree;


public class InternalParseTree<NT extends Enum<NT>> implements ParseTree<NT> {
    private final NT UNASSIGNED = null;
    
    /**
     * During the process of building the ParseTree, the name will sometimes be UNASSIGNED, 
     * but once the ParseTree is completed, the name should always be set to a NT. 
     */
    private NT name;
    private boolean isSkipped; // if this node is in a skipped subtree

    private final String contents; // the substring that this subtree matched
    private final int offset; // offset where this subtree matched in the original string
    private final Map<NT, List<ParseTree<NT>>> childrenByName;
    private final List<ParseTree<NT>> allChildren; // including skipped subtrees
    private final List<ParseTree<NT>> children;
    
    private void checkRep() {
        assert contents != null;
        assert childrenByName != null;
        assert allChildren != null;
    }

    InternalParseTree(String contents, int offset) {
        this(contents, Collections.emptyMap(), Collections.emptyList(), offset);
        checkRep();
    }
    
    InternalParseTree(String contents, Map<NT, List<ParseTree<NT>>> childrenByName, List<ParseTree<NT>> allChildren, int offset) {
        super();
        
        this.name = UNASSIGNED;
        this.isSkipped = false;

        this.contents = contents;
        this.offset = offset;
        this.childrenByName = childrenByName;
        this.allChildren = allChildren;
        
        List<ParseTree<NT>> theChildren = new ArrayList<>();
        for (ParseTree<NT> child : allChildren) {
            if (!child.isSkipped()) {
                theChildren.add(child);
            }
        }
        this.children = theChildren;
        
        checkRep();
    }
    
    InternalParseTree(NT nonterminal, String contents, Map<NT, List<ParseTree<NT>>> childrenByName, List<ParseTree<NT>> allChildren, int offset) {
        this(contents, childrenByName, allChildren, offset);
        assert nonterminal != null;
        this.name = nonterminal;
        checkRep();
    }

    @Override
    public NT name() {
        assert name != UNASSIGNED;
        return name;
    }
    
    @Override
    public String text(){
        return contents;
    }
    
    @Override
    public int start() {
        return offset;
    }

    @Override
    public int end() {
        return offset + contents.length();
    }
    
    @Override
    public List<ParseTree<NT>> children() {
        return Collections.unmodifiableList(children);
    }

    @Override
    public List<ParseTree<NT>> allChildren(){
        return Collections.unmodifiableList(allChildren);
    }
    
    @Override
    public boolean isSkipped() {
        return isSkipped;
    }
    
    @Override
    public List<ParseTree<NT>> childrenByName(NT nonterminalName){
        return childrenByName.containsKey(nonterminalName)
                ? Collections.unmodifiableList(childrenByName.get(nonterminalName))
                : Collections.emptyList();
    }
    
    @Override
    public String toString(){
        String s = (isSkipped ? "@skip " : "") + name;
        if (this.children().isEmpty()) {
            s += ":" + Display.escapeForReading(contents, "\"");
        } else {
            String t = "";
            int offsetReachedSoFar = start();
            for (ParseTree<NT> pt : allChildren){
                if (offsetReachedSoFar < pt.start()) {
                    // previous child and current child have a gap between them that must have been matched as a terminal
                    // in the rule for this node.  Insert it as a quoted string.
                    String terminal = contents.substring(offsetReachedSoFar - start(), pt.start() - start());
                    t += "\n" + Display.escapeForReading(terminal, "\"");
                }
                t += "\n" + pt;
                offsetReachedSoFar = pt.end();
            }
            if (offsetReachedSoFar < end()) {
                // final child and end of this node have a gap -- treat it the same as above.
                String terminal = contents.substring(offsetReachedSoFar - start());
                t += "\n" + Display.escapeForReading(terminal, "\"");                
            }
            final int smallEnoughForOneLine = 50;
            if (t.length() <= smallEnoughForOneLine) {
                s += " { " + t.substring(1) // remove initial newline
                              .replace("\n", ", ") 
                     + " }";
            } else {
                s += " {" + Display.indent(t, "  ") + "\n}";
            }
        }
        return s;
    }
    
    
    // Internal mutators used only while parse tree is being constructed.
    // Access level is package-private because other classes in the internal
    // package need to use these mutators.
    void setName(NT nonterminal) {
        assert nonterminal != UNASSIGNED;
        assert this.name == UNASSIGNED;
        this.name = nonterminal;
    }
    
    boolean assignedNT(){
        return name != UNASSIGNED;
    } 
    
    void markAsSkipped() {        
        this.isSkipped = true;
        for (ParseTree<NT> child : allChildren) {
            ((InternalParseTree<NT>) child).markAsSkipped();
        }
    }
        
    static <NT extends Enum<NT> > InternalParseTree<NT> concat(InternalParseTree<NT> p1, InternalParseTree<NT> p2){
        String newcontents = p1.contents + p2.contents;
        if(p1.isPlainString() && p2.isPlainString()){
            return new InternalParseTree<NT>(newcontents, p1.offset);
        }
        
        if(p1.assignedNT() && p2.assignedNT()){
            Map<NT, List<ParseTree<NT>>> newch = new HashMap<>();
            List<ParseTree<NT>> ch1 = new ArrayList<>();
            ch1.add(p1);
            newch.put(p1.name, ch1);
            if(p2.name.equals(p1.name)){
                ch1.add(p2);
            }else{
                List<ParseTree<NT>> ch2 = new ArrayList<>();
                ch2.add(p2);
                newch.put(p2.name, ch2);
            }     
            List<ParseTree<NT>> inorder = new ArrayList<>();
            inorder.add(p1);
            inorder.add(p2);
            return new InternalParseTree<NT>(newcontents, newch, inorder, p1.offset);
        }
        if(!p1.assignedNT() && p2.assignedNT()){
            if(p1.isTerminal()){
                Map<NT, List<ParseTree<NT>>> newch = new HashMap<>();
                List<ParseTree<NT>> ch1 = new ArrayList<>();
                ch1.add(p2);
                newch.put(p2.name, ch1);
                return new InternalParseTree<NT>(newcontents, newch, new ArrayList<>(ch1), p1.offset);
            }else{
                List<ParseTree<NT>> ch1;
                if(p1.childrenByName.containsKey(p2.name)){
                    ch1 = p1.childrenByName.get(p2.name);
                    ch1.add(p2);
                }else{
                    ch1 = new ArrayList<>();
                    ch1.add(p2);
                    p1.childrenByName.put(p2.name, ch1);                    
                }      
                p1.allChildren.add(p2);
                return new InternalParseTree<NT>(newcontents, p1.childrenByName, p1.allChildren, p1.offset);
            }
        }
        
        if(p1.assignedNT() && !p2.assignedNT()){
            if(p2.isTerminal()){
                Map<NT, List<ParseTree<NT>>> newch = new HashMap<>();
                List<ParseTree<NT>> ch1 = new ArrayList<>();
                ch1.add(p1);
                newch.put(p1.name, ch1);
                return new InternalParseTree<NT>(newcontents, newch, new ArrayList<>(ch1), p1.offset);
            }else{
                List<ParseTree<NT>> ch2;
                if(p2.childrenByName.containsKey(p1.name)){
                    ch2 = p2.childrenByName.get(p1.name);
                    ch2.add(0, p1);
                }else{
                    ch2 = new ArrayList<>();
                    ch2.add(p1);
                    p2.childrenByName.put(p1.name, ch2);                    
                }          
                p2.allChildren.add(0, p1);
                return new InternalParseTree<NT>(newcontents, p2.childrenByName, p2.allChildren, p1.offset);
            }
        }
        
        if(!p1.assignedNT() && !p2.assignedNT()){
            if(p2.isTerminal()){
                assert !p1.isTerminal();                
                return new InternalParseTree<NT>(newcontents, p1.childrenByName, p1.allChildren, p1.offset);
            }
            if(p1.isTerminal()){
                assert !p2.isTerminal();                
                return new InternalParseTree<NT>(newcontents, p2.childrenByName, p2.allChildren, p1.offset);
            }
            
            for(Entry<NT, List<ParseTree<NT>>> ent : p1.childrenByName.entrySet()){
                if(p2.childrenByName.containsKey(ent.getKey())){
                    ent.getValue().addAll(p2.childrenByName.remove(ent.getKey()));                    
                }
            }
            for(Entry<NT, List<ParseTree<NT>>> ent : p2.childrenByName.entrySet()){
                p1.childrenByName.put(ent.getKey(), ent.getValue());
            }
            p1.allChildren.addAll(p2.allChildren);
            return new InternalParseTree<NT>(newcontents, p1.childrenByName, p1.allChildren, p1.offset);
        }
       
        throw new AssertionError("should never get here");
    }

    private boolean isPlainString(){
        return name == UNASSIGNED && allChildren.isEmpty();
    }
    
    private boolean isTerminal() {
        return allChildren.isEmpty();
    }
    
}
