/* Copyright (c) 2016-2020 MIT 6.005/6.031 course staff, all rights reserved.
 * Redistribution of original or derived work requires permission of course staff.
 */
package edu.mit.eecs.parserlib.internal;

import java.io.IOException;
import java.io.Reader;
import java.util.Map;
import java.util.Map.Entry;

import edu.mit.eecs.parserlib.ParseTree;
import edu.mit.eecs.parserlib.Parser;
import edu.mit.eecs.parserlib.UnableToParseException;


public class InternalParser<NT extends Enum<NT>> implements Parser<NT> {
    private final Map<NT, GrammarTerm<NT>> grammar;
    private final NT start;
    
    /**
     * Make parser.
     * @param grammar grammar productions, mapping each nonterminal to its grammar expression
     * @param start starting nonterminal
     * @throws UnableToParseException thrown if productions have any undefined nonterminals
     */
    @SuppressWarnings("unchecked")
    public InternalParser(Map<NT, GrammarTerm<NT>> grammar, NT start) throws UnableToParseException {
        for(Object e : java.util.EnumSet.allOf(start.getClass())){            
            if(!grammar.containsKey(e)){
                throw new GrammarException("grammar is missing a definition for nonterminal " + e + " which is found in the enum type");
            }
        }
        this.grammar = grammar;
        this.start = start;
    }
    
    /**
     * Make parser from a grammar construct assembled by parsing a grammar.
     * @param gc grammar construct
     * @param start starting nonterminal
     * @throws UnableToParseException thrown if productions have any undefined nonterminals
     */
    public InternalParser(GrammarConstruct<NT> gc, NT start) throws UnableToParseException {
        this(gc.get(), start);
    }

    @Override
    public ParseTree<NT> parse(String textToParse) throws UnableToParseException {
        ParseResult<NT> pr;
        try{
            pr = new NonTerminal<NT>(start).parse(textToParse, 0, grammar, new ParserState<NT>());
        } catch (ParseException e) {
            // rethrow the exception, augmented by the original text, so that the error message is better
            throw new ParseException("string does not match grammar", e.nonterminalName, e.expectedText, textToParse, e.pos);
        }
        if(pr.failed()){            
            throw new ParseException("string does not match grammar", pr.nonterminal.toString(), pr.expectedText, textToParse, pr.pos);
        }
        if(pr.pos < textToParse.length()) {
            NT nonterminal = pr.lastFailure != null ? pr.lastFailure.nonterminal : start;
            String expectedText = pr.lastFailure != null ? pr.lastFailure.expectedText : "end of string";
            int pos = pr.lastFailure != null ? pr.lastFailure.pos : pr.pos;
            throw new ParseException("only part of the string matches the grammar; the rest did not parse", 
                                     nonterminal.toString(), expectedText, textToParse, pos);
        }
        
        return pr.tree;
    }

    @Override
    public ParseTree<NT> parse(Reader in) throws UnableToParseException, IOException{
        final int bufferSize = 10_000;
        char[] buffer = new char[bufferSize];
        StringBuffer sb = new StringBuffer();
        int n = 0;
        while ((n = in.read(buffer)) != -1) {
            sb.append(buffer, 0, n);
        }
        return parse(sb.toString());        
    }
    
    @Override
    public String toString(){
        String s = "";
        for(Entry<NT, GrammarTerm<NT>> ent : grammar.entrySet()){
            s += ent.getKey() + "::=" + ent.getValue() + ";\n";
        }
        return s;
    }
}
