/* Copyright (c) 2016-2020 MIT 6.005/6.031 course staff, all rights reserved.
 * Redistribution of original or derived work requires permission of course staff.
 */
package edu.mit.eecs.parserlib.internal;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import edu.mit.eecs.parserlib.ParseTree;
import edu.mit.eecs.parserlib.UnableToParseException;


/**
 * This class represents a nonterminal in the grammar. 
 * The type parameter <code>NT</code> is an Enum type that represents the set of nonterminals available in the grammar.
 * @author asolar
 *
 * @param <NT>
 */
public class NonTerminal<NT extends Enum<NT>> extends GrammarTerm<NT> {
    
    /**
     * 
     * @param name note that the type system prevents us from passing an illegal name.
     */
    public NonTerminal(NT name) {
        super();
        this.name = name;
    }

    private final NT name;

    @Override
    public ParseResult<NT> parse(String s, int pos, Map<NT, GrammarTerm<NT>> definitions, ParserState<NT> state) throws UnableToParseException {
        GrammarTerm<NT> gt = definitions.get(name);
        if(gt == null){
            // GrammarConstruct should be enforcing this invariant 
            throw new GrammarException("nonterminal has no definition: " + name);
        }
        state.enter(pos, name);
        ParseResult<NT> pr = gt.parse(s, pos, definitions, state);
        if (pr.failed()) {
            state.leave(name);
            return pr;
        }
        if(!pr.tree.assignedNT()){
            pr.tree.setName(this.name);
        }else{
            Map<NT, List<ParseTree<NT>>> chmap = new HashMap<>();
            List<ParseTree<NT>> alist = new ArrayList<>();
            alist.add(pr.tree);
            chmap.put(pr.tree.name(), alist);
            pr = pr.replaceTree( new InternalParseTree<NT>(this.name, pr.tree.text(), chmap, new ArrayList<>(alist), pr.tree.start())  );            
        }
        state.leave(name);
        return pr;
    }
    
    @Override
    public String toString(){
        return name.toString().toLowerCase();
    }
}
