/* Copyright (c) 2016-2020 MIT 6.005/6.031 course staff, all rights reserved.
 * Redistribution of original or derived work requires permission of course staff.
 */
package edu.mit.eecs.parserlib.internal;

/**
 * Immutable class that represents the result of trying to parse part of a string.
 * @author asolar
 *
 * @param <NT>
 */
public class ParseResult<NT extends Enum<NT>> {
    public enum Result {PASS, FAIL}
    public final Result res;   
    public final InternalParseTree<NT> tree;
    public final int pos;

    public final NT nonterminal;
    public final String expectedText;
    public final ParseResult<NT> lastFailure; // failing sub-parse that stopped this parse tree from being bigger
    // rep invariant
    //    res == PASS => nonterminal,expectedText == null; lastFailure == null || lastFailure.res == FAIL
    //    res == FAIL => nonterminal,expectedText != null; lastFailure == null
    //    

    private void checkRep() {
        switch (res) {
        case PASS:
            assert nonterminal == null;
            assert expectedText == null;
            assert lastFailure == null || lastFailure.res == Result.FAIL;
            break;
        case FAIL:
            assert nonterminal != null;
            assert expectedText != null;
            assert lastFailure == null;
            break;
        }
    }
    
    /**
     * Make a successful parse result.
     * @param tree parse tree produced by parse
     * @param pos offset in original string where the parse tree ends
     */
    public ParseResult(InternalParseTree<NT> tree, int pos) {
        this.res = Result.PASS;
        this.tree = tree;
        this.pos = pos;

        this.nonterminal = null;
        this.expectedText = null;
        this.lastFailure = null;
        checkRep();
    }
    
    /**
     * Make a failing parse result.
     * @param expectedText human-readable description of the text that the parser was looking for,
     *        e.g. ";" for a literal string, "[abc]" for a char class, "1|2|3" for a choice
     * @param pos offset in original string where parse failed
     */
    public ParseResult(NT nonterminal, String expectedText, int pos) {
        this.pos = pos;
        this.res = Result.FAIL;
        this.nonterminal = nonterminal;
        this.expectedText = expectedText;
        this.tree = null;
        this.lastFailure = null;
        checkRep();
    }


    /**
     * Keep track of a failing parse result that prevented this tree from matching more of the input string.
     * This deeper failure is usually more informative to the user, so we'll display it in the error message.
     * @param newLastFailure a failing ParseResult that stopped this tree's parse (but didn't prevent this from succeeding)
     * @return a new ParseResult identical to this one but with lastFailure added to it
     */
    public ParseResult<NT> addLastFailure(ParseResult<NT> newLastFailure) {
        assert newLastFailure.res == Result.FAIL;
        return new ParseResult<NT>(this.res, this.tree, this.pos, this.nonterminal, this.expectedText, 
                                   laterFailure(this.lastFailure, newLastFailure));
    }
    
    
    /**
     * @param newTree
     * @return a new ParseResult identical to this one but with newTree substituted as its parse tree
     */
    public ParseResult<NT> replaceTree(InternalParseTree<NT> newTree){
        assert this.res == Result.PASS;
        return new ParseResult<NT>(this.res, newTree, this.pos, this.nonterminal, this.expectedText, this.lastFailure);
    }
    
    /**
     * Concatenate this parse result with a parse result that immediately followed it. 
     * @param pr parse result immediately following this
     * @return a new ParseResult representing both successful parses, merging their parse trees
     */
    public ParseResult<NT> mergeResult(ParseResult<NT> pr){
        assert pr.res == Result.PASS && this.res == Result.PASS;
        return new ParseResult<NT>(this.res, InternalParseTree.concat(this.tree, pr.tree), 
                                   pr.pos, 
                                   this.nonterminal, 
                                   this.expectedText,
                                   laterFailure(this.lastFailure, pr.lastFailure));        
    }
    
    private static <NT extends Enum<NT>> ParseResult<NT> laterFailure(ParseResult<NT> failure1, ParseResult<NT> failure2) {
        if (failure1 == null) return failure2;
        else if (failure2 == null) return failure1;
        else if (failure1.pos >= failure2.pos) return failure1;
        else return failure2;
    }
    
    // private rep constructor, used by producers above
    private ParseResult(Result res, InternalParseTree<NT> tree, int pos, NT nonterminal, String expectedText, ParseResult<NT> lastFailure) {        
        this.res = res;
        this.tree = tree;
        this.pos = pos;
        this.nonterminal = nonterminal;
        this.expectedText = expectedText;
        this.lastFailure = lastFailure;
        checkRep();
    }
    
    /**
     * Test if this is a failing parse result.
     * @return true iff this represents a failing parse.
     */
    public boolean failed(){
        return res == Result.FAIL;
    }
    
    @Override
    public String toString() {
        return (res == Result.PASS)
                ? res + "," + pos + "," + tree + (lastFailure != null ? ",[" + lastFailure + "]" : "")
                : res + "," + pos + "," + nonterminal + "," + Display.escapeForReading(expectedText, "");
    }
}
