/* Copyright (c) 2016-2020 MIT 6.005/6.031 course staff, all rights reserved.
 * Redistribution of original or derived work requires permission of course staff.
 */
package edu.mit.eecs.parserlib.internal;

import java.util.Map;

import edu.mit.eecs.parserlib.UnableToParseException;


/**
 * This class represents different kinds of repetitions in the grammar. It can either represent
 * (a) a Kleen closure (x*) which matches zero or more instances of a term <code>x</code>, 
 * (b) the operator x+ which matches one or more instances of <code>x</code>, or 
 * (c) the operator x? which matches zero or one copies of <code>x</code>.
 * 
 * The type parameter <code>NT</code> is an Enum type that represents the set of nonterminals available in the grammar.
 * @author asolar
 *
 * @param <NT>
 */
public class Repetition<NT extends Enum<NT>> extends GrammarTerm<NT> {
    public static interface HowMany { 
        public static HowMany ZERO_OR_MORE = new AtLeast(0);
        public static HowMany ONE_OR_MORE = new AtLeast(1);
        public static HowMany ZERO_OR_ONE = new Between(0, 1);
        
        public boolean tooLow(int n);
        public boolean tooHigh(int n);
    }
    
    public static class AtLeast implements HowMany {
        private final int min; // min >= 0
        public AtLeast(int min) { this.min = min; }
        public boolean tooLow(int n) { return n < min; }
        public boolean tooHigh(int n) { return false; }
        public String toString() {
            switch (min) {
            case 0: return "*";
            case 1: return "+";
            default: return "{" + min + ",}";
            }
        }
    }
    
    public static class Between implements HowMany {
        private final int min; // min >= 0
        private final int max; // max >= min
        public boolean tooLow(int n) { return n < min; }
        public boolean tooHigh(int n) { return n > max; }
        public Between(int min, int max) { this.min = min; this.max = max; }
        public String toString() {
            if (min == 0) { return (max == 1) ? "?" : "{," + max + "}"; }
            else { return "{" + min + "," + max + "}"; }
        }
    }
        
    
    private final GrammarTerm<NT> body;
    private final HowMany howmany;
    
    /**
     * Construct a kind of repetition determined by the <code>howmany</code> parameter.
     * @param body of this repetition
     * @param howmany number of occurrences expected
     */
    public Repetition(GrammarTerm<NT> body, HowMany howmany) {
        super();
        this.howmany = howmany;
        this.body = body;
    }
    
    @Override
    public ParseResult<NT> parse(String s, int pos, Map<NT, GrammarTerm<NT>> definitions, ParserState<NT> state) throws UnableToParseException {
        ParseResult<NT> prout=null;
        for (int timesMatched = 0; howmany.tooLow(timesMatched) || !howmany.tooHigh(timesMatched+1); ++timesMatched){
            ParseResult<NT> pr = body.parse(s, pos, definitions, state);
            if (pr.failed()){
                // no match
                if(howmany.tooLow(timesMatched)) {
                    return pr; // return the failure; FIX needs to do right thing for {n,m}
                }
                if (prout == null){
                    prout = new ParseResult<NT>(new InternalParseTree<NT>("", pos), pos);
                } else {
                    prout = prout.addLastFailure(pr);
                }
                return prout;
            } else if (pr.pos == pos){
                // matched the empty string.  we'll get into an infinite loop if we repeat the loop again, so
                // don't; but on the positive side, we would be able to repeat this as many times as we want, 
                // so timesMatched can definitely satisfy howmany. return successful match at this point
                return (prout == null) ? pr : prout;
            } else {
                // matched a nonempty string; advance the position and merge pr into prout
                pos = pr.pos;
                prout = (prout == null) ? pr : prout.mergeResult(pr);
            }
        }
        return (prout == null) ? new ParseResult<NT>(new InternalParseTree<NT>("", pos), pos) : prout;
    }

    @Override
    public String toString(){
        return body.toString() + howmany.toString();
    }
    
}
