import assert from 'assert';
import { Parser, ParseTree, InternalParseError, GrammarError } from './types';
import { InternalParseTree } from './parsetree';

/**
 * Defines a term in a grammar, providing a parse method that parses that particular kind of term.
 */
export interface GrammarTerm<NT> {

    /**
     * Given a string s and a position pos within the string, this method will attempt to parse 
     * this particular GrammarTerm<NT> from the string. If the GrammarTerm<NT> is successfully parsed, 
     * the ParseResult<NT> will include a {@link InternalParseTree<NT>} for the term and the position in the string where
     * that term ends. This means that {@link ParseResult<NT>}.pos is between pos and s.length.
     * 
     * If the GrammarTerm<NT> is not parsed successfully, then the ParseResult<NT> will
     * indicate that (see definition of {@link ParseResult<NT>}). 
     *  
     * 
     * @param s
     * @param pos should be a valid position in the string s.
     * @param definitions should contain a definition of *all* the nonterminals in the grammar.
     * @return result of matching this grammar term against s starting from position pos
     * @throws UnableToParseException if a parse failure occurs that can't be fixed by backtracking
     */
    parse(s:string, pos:number, definitions:GrammarDefinitions<NT>, state:ParserState<NT>):ParseResult<NT>;

    /**
     * @return grammar expression for this term
     */
    toString():string;
}


 /**
 * A particular use of the GrammarConstruct class overrides this method to actually return 
 * a map of GrammarTerm<NT>s.
 * Requires every key to be a lower-case nonterminal.
 * @return Map of nonterminals and their corresponding grammar terms. All elements in the NT Enum must
 * be represented in this map.
 * @throws UnableToParseException if the nonterminals in the grammar are not in 1-1 correspondence with the symbols in NT
 */
export type GrammarDefinitions<NT> = Map<NT, GrammarTerm<NT>>;


export function nt<NT>(nonterminal:NT, nonterminalName:string):GrammarTerm<NT> {
    return {
        parse(s, pos, definitions, state) {
            const gt = definitions.get(nonterminal);
            if (gt === undefined) throw new GrammarError("nonterminal has no definition: " + nonterminalName);
            // console.error("entering", nonterminalName);
            state.enter(pos, nonterminal);
            let pr = gt.parse(s, pos, definitions, state);
            state.leave(nonterminal);
            // console.error("leaving", nonterminalName, "with result", pr);
            if (!pr.failed && !state.isEmpty()) {
                const tree = pr.tree!;
                const newTree = state.makeParseTree(tree.start, tree.text, [tree]);
                pr = pr.replaceTree(newTree);
            }
            return pr;
        },
        
        toString() {
            return nonterminalName;
        }
    }
}

export function regex<NT>(regexSource:string):GrammarTerm<NT> {
    let regex = new RegExp('^' + regexSource + '$', 's');
    return {
        parse(s, pos, definitions, state) {
            if (pos >= s.length) {
                return state.makeFailedParse(pos, regexSource);
            }
            const l = s.substring(pos, pos+1);
            if (regex.test(l)) {
                return state.makeSuccessfulParse(pos, pos+1, l);
            }
            return state.makeFailedParse(pos, regexSource);
        },
        
        toString() {
            return regexSource;
        }
    }
}

export function str<NT>(str:string):GrammarTerm<NT> {
    return {
        parse(s, pos, definitions, state) {
            const newpos = pos + str.length;
            if (newpos > s.length) {
                return state.makeFailedParse(pos, str);
            }
            const l = s.substring(pos, newpos);
            if (l === str) {
                return state.makeSuccessfulParse(pos, newpos, l);
            }
            return state.makeFailedParse(pos, str);
        },
        
        toString() {
            return "'" + str.replace(/'\r\n\t\\/, "\\$&") + "'";
        }
    }
}

export function cat<NT>(...terms:GrammarTerm<NT>[]):GrammarTerm<NT> {
    return {
        parse(s, pos, definitions, state) {
            let prout = state.makeSuccessfulParse(pos, pos);
            for (const gt of terms) {
                const pr = gt.parse(s, pos, definitions, state);
                if (pr.failed) return pr;
                pos = pr.pos;
                prout = prout.mergeResult(pr);
            }
            return prout;
        },
        
        toString() {
            return "(" + terms.map(term => term.toString()).join(" ") + ")";
        }
    }
}

/**
 * @param choices must be nonempty
 */
export function or<NT>(...choices:GrammarTerm<NT>[]):GrammarTerm<NT> {
    assert(choices.length > 0);
    return {
        parse(s, pos, definitions, state) {
            const successes:SuccessfulParse<NT>[] = []; 
            const failures:FailedParse<NT>[] = [];
            choices.forEach((choice) => {
                const result = choice.parse(s, pos, definitions, state);
                if (result.failed) {
                    failures.push(result);
                } else {
                    successes.push(result);
                }
            });
            if (successes.length > 0) {
                const longestSuccesses = longestResults(successes);
                assert(longestSuccesses.length > 0);
                return longestSuccesses[0];
            }
            const longestFailures = longestResults(failures);
            assert(longestFailures.length > 0);
            return state.makeFailedParse(longestFailures[0].pos, longestFailures.map((result:FailedParse<NT>) => result.expectedText).join("|"));
        },
        
        toString() {
            return "(" + choices.map(choice => choice.toString()).join("|") + ")";
        }
    }
}


export interface HowMany {
    tooLow(n:number):boolean;
    tooHigh(n:number):boolean;
    toString():string;
}
export class AtLeast implements HowMany {
    constructor(private readonly min:number) {}
    tooLow(n:number):boolean { return n < this.min; }
    tooHigh(n:number):boolean { return false; }
    toString():string {
        switch (this.min) {
        case 0: return "*";
        case 1: return "+";
        default: return "{" + this.min + ",}";
        }
    }
}
export class Between implements HowMany {
    constructor(
        private readonly min:number,
        private readonly max:number
    ) {}
    tooLow(n:number):boolean { return n < this.min; }
    tooHigh(n:number):boolean { return n > this.max; }
    toString():string {
        if (this.min == 0) { return (this.max == 1) ? "?" : "{," + this.max + "}"; }
        else { return "{" + this.min + "," + this.max + "}"; }
    }
}
export const ZERO_OR_MORE:HowMany = new AtLeast(0);
export const ONE_OR_MORE:HowMany = new AtLeast(1);
export const ZERO_OR_ONE:HowMany = new Between(0, 1);

export function repeat<NT>(gt:GrammarTerm<NT>, howmany:HowMany):GrammarTerm<NT> {
    return {
        parse(s, pos, definitions, state) {
            let prout = state.makeSuccessfulParse(pos, pos);
            for (let timesMatched = 0; howmany.tooLow(timesMatched) || !howmany.tooHigh(timesMatched+1); ++timesMatched) {
                const pr = gt.parse(s, pos, definitions, state);
                if (pr.failed) {
                    // no match
                    if (howmany.tooLow(timesMatched)) {
                        return pr;
                    }
                    return prout.addLastFailure(pr);
                } else {
                    if (pr.pos == pos) {
                        // matched the empty string, and we already have enough.
                        // we may get into an infinite loop if howmany.tooHigh() never returns false,
                        // so return successful match at this point
                        return prout;
                    }
                    // otherwise advance the position and merge pr into prout
                    pos = pr.pos;
                    prout = prout.mergeResult(pr);
                }
            }
            return prout;
        },
        
        toString() {
            return gt.toString() + howmany.toString();
        }
    };
}

export function star<NT>(gt:GrammarTerm<NT>):GrammarTerm<NT> {
    return repeat(gt, ZERO_OR_MORE);
}

export function plus<NT>(gt:GrammarTerm<NT>):GrammarTerm<NT> {
    return repeat(gt, ONE_OR_MORE);
}

export function option<NT>(gt:GrammarTerm<NT>):GrammarTerm<NT> {
    return repeat(gt, ZERO_OR_ONE);
}

export function skip<NT>(nonterminal:GrammarTerm<NT>):GrammarTerm<NT> {
    const repetition = star(nonterminal);
    return {
        parse(s, pos, definitions, state) {
            state.enterSkip();
            let pr = repetition.parse(s, pos, definitions, state);
            state.leaveSkip();
            if (pr.failed) {
                // succeed anyway
                pr = state.makeSuccessfulParse(pos, pos);
            }
            return pr;
        },
        
        toString() {
            return "(?<skip>" + repetition + ")";
        }
    };
}

export function failfast<NT>(gt:GrammarTerm<NT>):GrammarTerm<NT> {
    return {
        parse(s, pos, definitions, state) {
            let pr = gt.parse(s, pos, definitions, state);
            if (pr.failed) throw new InternalParseError("", pr.nonterminalName!, pr.expectedText!, "", pr.pos);
            return pr;
        },

        toString() {
            return 'failfast(' + gt + ')';
        }
    };
}


export class InternalParser<NT> implements Parser<NT> {
    constructor(
        private readonly definitions:GrammarDefinitions<NT>,
        private readonly start:GrammarTerm<NT>,
        private nonterminalToString:(nt:NT)=>string
    ) {
        this.checkRep();
    }

    private checkRep() {
    }

    parse(textToParse:string):ParseTree<NT> {
        let pr = (() => {
            try {
                return this.start.parse(textToParse, 0, this.definitions, new ParserState<NT>(this.nonterminalToString));
            } catch (e) {
                if (e instanceof InternalParseError) {
                    // rethrow the exception, augmented by the original text, so that the error message is better
                    throw new InternalParseError("string does not match grammar", e.nonterminalName, e.expectedText, textToParse, e.pos);
                } else {
                    throw e;
                }
            }
        }) ();

        if (pr.failed) {
            throw new InternalParseError("string does not match grammar",
                                         pr.nonterminalName, pr.expectedText, textToParse, pr.pos);
        }

        if (pr.pos < textToParse.length) {
            const message = "only part of the string matches the grammar; the rest did not parse";
            throw (pr.lastFailure
                ? new InternalParseError(message, pr.lastFailure.nonterminalName, pr.lastFailure.expectedText, textToParse, pr.lastFailure.pos)
                : new InternalParseError(message, this.start.toString(), "end of string", textToParse, pr.pos)
            );
        }

        return pr.tree;
    };

    toString():string {
        return Array.from(this.definitions, ([nonterminal, rule]) => this.nonterminalToString(nonterminal) + '::=' + rule + ';').join("\n");
    }
}

// ParseResult is a discriminated union, discriminated on the failed property (true or false)
type ParseResult<NT> = SuccessfulParse<NT> | FailedParse<NT>;

export interface BaseResult<NT> {
    readonly failed:boolean; // did parse fail (FailedParse) or succeed (SuccessfulParse)
    readonly pos:number; // how far into the string the parse reached
}

export class SuccessfulParse<NT> implements BaseResult<NT>
{
    readonly failed:false = false;

    constructor(
        readonly pos:number,
        readonly tree:InternalParseTree<NT>,
        readonly lastFailure?:FailedParse<NT>
    ) {
    }

    replaceTree(tree:InternalParseTree<NT>):SuccessfulParse<NT> {
        return new SuccessfulParse(this.pos, tree, this.lastFailure);
    }

    mergeResult(that:SuccessfulParse<NT>):SuccessfulParse<NT> {
        assert(!that.failed);
        //console.log('merging', this, 'with', that);
        return new SuccessfulParse(
            that.pos,
            this.tree.concat(that.tree!),
            laterResult(this.lastFailure, that.lastFailure)
        );
    }

    /**
     * Keep track of a failing parse result that prevented this tree from matching more of the input string.
     * This deeper failure is usually more informative to the user, so we'll display it in the error message.
     * @param newLastFailure a failing ParseResult<NT> that stopped this tree's parse (but didn't prevent this from succeeding)
     * @return a new ParseResult<NT> identical to this one but with lastFailure added to it
     */
    addLastFailure(newLastFailure:FailedParse<NT>):SuccessfulParse<NT> {
        assert(newLastFailure.failed);
        return new SuccessfulParse(this.pos, this.tree, 
                                   laterResult(this.lastFailure, newLastFailure));
    }
    
}

export class FailedParse<NT> implements BaseResult<NT>
{
    readonly failed:true = true;

    constructor(
        readonly pos:number,
        readonly nonterminalName:string,
        readonly expectedText:string
    ) {
    }

}


/**
 * @param result1 
 * @param result2 
 * @return whichever of result1 or result2 has the mximum position, or undefined if both are undefined
 */
function laterResult<NT, ResultKind extends BaseResult<NT>>(result1:ResultKind|undefined, result2:ResultKind|undefined):ResultKind|undefined {
    if (result1 && result2) return result1.pos >= result2.pos ? result1 : result2;
    else return result1 || result2;
}


/**
 * @param results
 * @return the results in the list with maximum pos.  Empty if list is empty.
 */
function longestResults<NT, ResultKind extends ParseResult<NT>>(results:ResultKind[]):ResultKind[] {
    return results.reduce((longestResultsSoFar:ResultKind[], result:ResultKind) => {
        if (longestResultsSoFar.length == 0 || result.pos > longestResultsSoFar[0].pos) {
            // result wins
            return [result];
        } else if (result.pos == longestResultsSoFar[0].pos) {
            // result is tied
            longestResultsSoFar.push(result);
            return longestResultsSoFar;
        } else {
            // result loses
            return longestResultsSoFar;
        }
    }, []);
}



type Stack<T> = T[];

export class ParserState<NT> {

    private readonly stack:Stack<NT> = [];
    private readonly first:Map<NT, Stack<number>> = new Map();
    private skipDepth = 0;

    constructor(private readonly nonterminalToString:(nt:NT)=>string) {
    }

    enter(pos:number, nonterminal:NT):void {
        if (! this.first.has(nonterminal)) {
            this.first.set(nonterminal, []);
        }
        const s = this.first.get(nonterminal)!;
        if (s.length > 0 && s[s.length-1] == pos) {
            throw new GrammarError("detected left recursion in rule for " + this.nonterminalToString(nonterminal));
        }
        s.push(pos);
        this.stack.push(nonterminal);
    }

    leave(nonterminal:NT):void {
        assert(this.first.has(nonterminal) && this.first.get(nonterminal)!.length > 0);
        this.first.get(nonterminal)!.pop();
        const last = this.stack.pop();
        assert(last === nonterminal);
    }

    enterSkip() {
        //console.error('entering skip');
        ++this.skipDepth;
    }

    leaveSkip() {
        //console.error('leaving skip');
        --this.skipDepth;
        assert(this.skipDepth >= 0);
    }

    isEmpty():boolean {
        return this.stack.length == 0;
    }

    get currentNonterminal():NT|undefined {
        return this.stack[this.stack.length-1];
    }

    get currentNonterminalName():string|undefined {
        return this.currentNonterminal !== undefined ? this.nonterminalToString(this.currentNonterminal) : undefined;
    }

    // requires: !isEmpty()
    makeParseTree(pos:number, text:string = '', children:ParseTree<NT>[] = []):InternalParseTree<NT> {
        assert(!this.isEmpty());
        return new InternalParseTree<NT>(
            this.currentNonterminal!,
            this.currentNonterminalName!,
            pos,
            text,
            children,
            this.skipDepth > 0);
    }

    // requires !isEmpty()
    makeSuccessfulParse(fromPos:number, toPos:number, text:string = '', children:ParseTree<NT>[] = []):SuccessfulParse<NT> {
        assert(!this.isEmpty());
        return new SuccessfulParse(toPos, this.makeParseTree(fromPos, text, children));
    }

    // requires !isEmpty()
    makeFailedParse(atPos:number, expectedText:string):FailedParse<NT> {
        assert(!this.isEmpty());
        return new FailedParse(atPos, this.currentNonterminalName!, expectedText);
    }
}

