import assert from 'assert';
import { ParseTree } from "./types";
import { escapeForReading, indent } from "./display";

export class InternalParseTree<NT> implements ParseTree<NT> {

    constructor(
        readonly name:NT,
        private readonly nonterminalName:string,
        readonly start:number,
        readonly text:string,
        readonly allChildren:/*frozen*/ParseTree<NT>[],
        readonly isSkipped:boolean
    ) {
        this.checkRep();
        Object.freeze(this.allChildren);
        // can't freeze(this) because of beneficent mutation delayed computation-with-caching for children() and childrenByName()
    }

    private checkRep() {
        // FIXME
    }
    
    get end():number {
        return this.start + this.text.length;
    }

    private _childrenByName?:Map<NT, /*frozen*/ParseTree<NT>[]>;
    childrenByName(name:NT):/*frozen*/ParseTree<NT>[] {
        if (! this._childrenByName) {
            this._childrenByName = new Map();
            for (const child of this.allChildren) {
                if (! this._childrenByName.has(child.name)) {
                    this._childrenByName.set(child.name, []);
                }
                this._childrenByName.get(child.name)!.push(child);
            }
            for (const childList of this._childrenByName.values()) {
                Object.freeze(childList);
            }
        }
        this.checkRep();
        return this._childrenByName!.get(name) || [];
    }

    private _children?:/*frozen*/ParseTree<NT>[];
    get children():/*frozen*/ParseTree<NT>[] {
        if (! this._children) {
            this._children = this.allChildren.filter(child => !child.isSkipped);
            Object.freeze(this._children);
        }
        this.checkRep();
        return this._children;
    }

    concat(that:InternalParseTree<NT>):InternalParseTree<NT> {
        return new InternalParseTree<NT>(
            this.name,
            this.nonterminalName,
            this.start,
            this.text + that.text,
            this.allChildren.concat(that.allChildren),
            this.isSkipped && that.isSkipped
        );
    }

    toString() {
        let s = (this.isSkipped ? "@skip " : "") + this.nonterminalName;
        if (this.children.length == 0) {
            s += ":" + escapeForReading(this.text, "\"");
        } else {
            let t = "";
            let offsetReachedSoFar = this.start;
            for (const pt of this.allChildren){
                if (offsetReachedSoFar < pt.start) {
                    // previous child and current child have a gap between them that must have been matched as a terminal
                    // in the rule for this node.  Insert it as a quoted string.
                    const terminal = this.text.substring(offsetReachedSoFar - this.start, pt.start - this.start);
                    t += "\n" + escapeForReading(terminal, "\"");
                }
                t += "\n" + pt;
                offsetReachedSoFar = pt.end;
            }
            if (offsetReachedSoFar < this.end) {
                // final child and end of this node have a gap -- treat it the same as above.
                const terminal = this.text.substring(offsetReachedSoFar - this.start);
                t += "\n" + escapeForReading(terminal, "\"");                
            }
            const smallEnoughForOneLine = 50;
            if (t.length <= smallEnoughForOneLine) {
                s += " { " + t.substring(1) // remove initial newline
                              .replace("\n", ", ") 
                     + " }";
            } else {
                s += " {" + indent(t, "  ") + "\n}";
            }
        }
        return s;
    }
}
