import { ParseTree, NonterminalEnum } from "./types";
import { makeNonterminalConverters } from "./compiler";
import { VERSION } from "../parserlib";
import fs from 'fs';
import path from 'path';

// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Iteration_protocols#The_iterator_protocol
type Iterator<E> = { next():{done:boolean, value?:E} };

function emptyIterator<E>():Iterator<E> {
    return {
        next() { return { done: true}; }
    };
}

function getIterator<E>(list:E[]):Iterator<E> {
    return <Iterator<E>>list[Symbol.iterator]();
}

const MAX_URL_LENGTH_FOR_DESKTOP_BROWSE = 2020;

/**
 * Visualizes a parse tree using a URL that can be pasted into your web browser.
 * @param parseTree tree to visualize
 * @param <NT> the enumeration of symbols in the parse tree's grammar
 * @return url that shows a visualization of the parse tree
 */
export function visualizeAsUrl<NT>(parseTree:ParseTree<NT>, nonterminals:NonterminalEnum):string {
    const base = "http://6031.mit.edu/parserlib/" + VERSION + "/visualizer.html";
    const code = expressionForDisplay(parseTree, nonterminals);
    const url = base + '?code=' + fixedEncodeURIComponent(code);
    if (url.length > MAX_URL_LENGTH_FOR_DESKTOP_BROWSE) {
        // display alternate instructions to the console
        console.error(
            'Visualization URL is too long for web browser and/or web server.\n'
            + 'Instead, go to ' + base + '\n'
            + 'and copy and paste this code into the textbox:\n'
            + code);
    }
    return url;
}


const visualizerHtmlFile = path.resolve(__dirname, '../../src/visualizer.html');

/**
 * Visualizes a parse tree as a string of HTML that can be displayed in a web browser.
 * @param parseTree tree to visualize
 * @param <NT> the enumeration of symbols in the parse tree's grammar
 * @return string of HTML that shows a visualization of the parse tree
 */
export function visualizeAsHtml<NT>(parseTree:ParseTree<NT>, nonterminals:NonterminalEnum):string {
    const html = fs.readFileSync(visualizerHtmlFile, 'utf8');
    const code = expressionForDisplay(parseTree, nonterminals);
    const result = html.replace(/\/\/CODEHERE/, 
                    "return '" + fixedEncodeURIComponent(code) + "';");
    return result;
}

function expressionForDisplay<NT>(parseTree:ParseTree<NT>, nonterminals:NonterminalEnum) {
    const {nonterminalToString} = makeNonterminalConverters<NT>(nonterminals);
    return forDisplay(parseTree, [], parseTree);

    function forDisplay(node:ParseTree<NT>, siblings:ParseTree<NT>[], parent:ParseTree<NT>):string {
        const name = nonterminalToString(node.name).toLowerCase();
        let s = "nd(";
        if (node.children.length == 0) {
            s += "\"" + name + "\",nd(\"'"+ cleanString(node.text) + "'\"),";
        } else {
            s += "\"" + name + "\",";
            const children = node.allChildren.slice(); // make a copy for shifting
            const firstChild = children.shift()!;
            let childrenExpression = forDisplay(firstChild, children, node);
            if (node.start < firstChild.start) {
                // node and its first child have a gap between them that must have been matched as a terminal
                // in the rule for node.  Insert it as a quoted string.
                childrenExpression = precedeByTerminal(node.text.substring(0, firstChild.start - node.start), childrenExpression);
            }
            s += childrenExpression + ",";
        }
        if (siblings.length > 0) {
            const sibling = siblings.shift()!;
            let siblingExpression = forDisplay(sibling, siblings, parent);
            if (node.end < sibling.start) {
                // node and its sibling have a gap between them that must have been matched as a terminal
                // in the rule for parent.  Insert it as a quoted string.
                siblingExpression = precedeByTerminal(parent.text.substring(node.end - parent.start, sibling.start - parent.start), siblingExpression);
            }
            s += siblingExpression;
        } else {
            let siblingExpression = "uu";
            if (node.end < parent.end) {
                // There's a gap between the end of node and the end of its parent, which must be a terminal matched by parent.
                // Insert it as a quoted string.
                siblingExpression = precedeByTerminal(parent.text.substring(node.end - parent.start), siblingExpression);
            }
            s += siblingExpression;
        }
        if (node.isSkipped) {
            s += ",true";
        }
        s += ")";
        return s;
    }

    function precedeByTerminal(terminal:string, expression:string):string {
        return "nd(\"'"+ cleanString(terminal) + "'\", uu, " + expression + ")";     
    }

    function cleanString(s:string):string {
        let rvalue = s.replace("\\", "\\\\");
        rvalue = rvalue.replace("\"", "\\\"");
        rvalue = rvalue.replace("\n", "\\n");
        rvalue = rvalue.replace("\r", "\\r");
        return rvalue;
    }
    
}

// from https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/encodeURIComponent
function fixedEncodeURIComponent(s:string):string {
    return encodeURIComponent(s).replace(/[!'()*]/g, c => '%' + c.charCodeAt(0).toString(16));
}

